#' Thomas Johnson III
#' Function to get the CE data

library(MASS)
library(tidyverse)

get_CE_data = function(dat.seed = 100, n_males = 1000, n_females = 1000)
{
  
  # Load the data
  set.seed(dat.seed)
  # Using SALARYX as the sensitive y variable
  # Excluding the idenitification and heavily related and likely senstive financial variables.
  CE_xy = rpms::CE %>% dplyr::select(-c(NEWID, PSU, CID, QINTRVMO, STATE, INCNONWK, 
                                        ARM_FORC, IN_COLL, POPSIZE, CUTENURE, FAM_TYPE,MARITAL, EDUCA,
                                        MEMBRACE, EARNER, OCCUCODE, REGION, BLS_URBN, HORIGIN, FINLWT21)) %>%
    filter(is.na(SALARYX) == FALSE) %>%
    filter(FINCBTAX <= quantile(FINCBTAX, probs = 0.75) + 3 * IQR(FINCBTAX), 
           FINCBTAX >= quantile(FINCBTAX, probs = 0.25) -3 * IQR(FINCBTAX),
           TOTXEST <= quantile(TOTXEST, probs = 0.75) + 3 * IQR(TOTXEST), 
           TOTXEST >= quantile(TOTXEST, probs = 0.25) -3 * IQR(TOTXEST),
           TOTEXPCQ <= quantile(TOTEXPCQ, prob = 0.75) + 3 * IQR(TOTEXPCQ),
           TOTEXPCQ >= quantile(TOTEXPCQ, prob = 0.25) - 3 * IQR(TOTEXPCQ),
           SALARYX <= quantile(SALARYX, probs = .75) + 3 * IQR(SALARYX),
           SALARYX >= quantile(SALARYX, probs = .25) - 3 * IQR(SALARYX))%>%
    mutate(TOTXEST = TOTXEST + abs(min(TOTXEST)) + 0.0001,
           TOTEXPCQ = TOTEXPCQ + abs(min(TOTEXPCQ)) + 0.0001,
           FINCBTAX = FINCBTAX + abs(min(FINCBTAX)) + 0.0001) %>%
    dplyr::select("AGE", "SEX", "TOTXEST", "TOTEXPCQ", "FINCBTAX", "SALARYX")
  
  # Getting imputation values
  # mode_popsize = 1
  # avg_roomsq = as.integer(mean(CE_xy$ROOMSQ, na.rm = TRUE))
  # med_bedroomq = as.integer(median(CE_xy$BEDROOMQ, na.rm = TRUE))
  # med_bathroomq = as.integer(median(CE_xy$BATHRMQ, na.rm = TRUE))
  # mode_arm_forc = "N"
  # mode_in_coll = "No"
  # med_irax = as.integer(median(CE_xy$IRAX,na.rm = TRUE))
  # med_liquidx = as.integer(median(CE_xy$IRAX, na.rm = TRUE))
  # med_stockx = as.integer(median(CE_xy$STOCKX,na.rm = TRUE))
  # med_studentx = as.integer(median(CE_xy$STUDNTX, na.rm = TRUE))
  
  # Reference: https://stackoverflow.com/questions/39126537/replace-na-in-a-factor-column
  
  # CE_xy = CE_xy %>% replace_na(
  #   replace = list(
  #     ROOMSQ = avg_roomsq,
  #     BEDROOMQ = med_bedroomq,
  #     BATHRMQ = med_bathroomq,
  #     IRAX = med_irax,
  #     LIQUIDX = med_liquidx,
  #     STOCKX = med_stockx,
  #     STUDNTX = med_studentx)) %>% na.omit()
  
  # Response: SALARYX, FINCBTAX
  # ind. Variables: STOCKX, LIQUIDX, STUDNTX, IRAX, TOTXEST, SEX, EHOUSNGC
  
  # Divide into two groups based on male and female, then sample from both groups then combine
  set.seed(90)
  CE_xy_male = CE_xy %>% filter(SEX == "M")
  CE_xy_female = CE_xy %>% filter(SEX == "F")
  CE_xy_dat = rbind(CE_xy_male[sample(1:nrow(CE_xy_male), n_males, replace = FALSE),],
                    CE_xy_female[sample(1:nrow(CE_xy_female), n_females, replace = FALSE),])
  
  # Rewrite the SALARYX and FINCBTAX to y_1, y_2
  variable_rewrite = function(CE_dat)
  {
    CE_dat = CE_dat %>% rename(y_2 = SALARYX, y_1 = FINCBTAX)
    CE_dat = model.matrix(~., data = CE_dat)[,-1]
    CE_dat = data.frame(CE_dat)
    return(CE_dat)
  }
  # Perform variable rewrite
  CE_xy_dat_preproc = variable_rewrite(CE_dat = CE_xy_dat)
  # shift to prevent TOTXEST from being negative
  CE_xy_dat_preproc = CE_xy_dat_preproc %>% 
    mutate(TOTXEST = TOTXEST + abs(min(TOTXEST)) + .Machine$double.eps)
  
  # put dataframes into a list
  return(CE_xy_dat_preproc)
}

readr::write_csv(get_CE_data(),paste0(here::here(),"/Data/cleaned-CE.csv"))
