#' Thoams Johnson III
#' Code for perfroming GADP and CGADP
#' Note that the CGADP requires a specific script for each model at the moment


# Model 1 GADP ------

gadp_func = function(unprotected_data, target_indices, theta = 0.4, dat.seed = 12)
{
  set.seed(dat.seed)
  
  unprotected_data = as.matrix(unprotected_data)
  # means of the target variable
  target_means = colMeans(unprotected_data[,target_indices])
  # means of the auxiliary variable
  auxiliary_means = colMeans(unprotected_data[,-target_indices])
  # Covariance of the target variable
  target_covariance = cov(unprotected_data)[target_indices,target_indices]
  
  
  # Auxiliary variables covariance
  aux_covariance = cov(unprotected_data[,-target_indices])
  
  target_aux_covariance = as.matrix(Matrix::nearPD(cov(unprotected_data))$mat)
  target_aux_covariance = 
    target_aux_covariance[-c(target_indices),
                          c(target_indices)]
  
  # print("auxiliary covariance matrix")
  # print(dim(target_aux_covariance))
  # 
  # print("target covariance matrix")
  # print(dim(target_covariance))
  
  unprotected_cov = cov(unprotected_data)
  
  unprotected_means = colMeans(unprotected_data)
  
  
  
  # random_variates = MASS::mvrnorm(mu = c(target_means, auxiliary_means),
  #                                 Sigma = )
  
  new_y_variates_covariance = unprotected_cov[target_indices,target_indices] * theta^2
  new_y_variates_covariance = cbind(unprotected_cov[target_indices,-target_indices] , 
                                    new_y_variates_covariance)
  
  # print("Randomized y covariance dimensions")
  # print(dim(new_y_variates_covariance))
  # 
  # print("target covariance")
  # print(target_covariance)
  # print(dim(target_covariance))
  # print("covariance for new variables")
  # print(new_y_variates_covariance)
  # print("covarinace for unprotected variables")
  # print(unprotected_cov)
  
  right_term = new_y_variates_covariance %*%
    solve(unprotected_cov) %*% t(new_y_variates_covariance)
  
  right_term = as.matrix(Matrix::nearPD(right_term)$mat)
  
  # print(right_term)
  # 
  # print("check positive definite")
  #   print(matrixcalc::is.positive.definite(right_term))
  #   print(matrixcalc::is.positive.definite(target_covariance))

  
  new_covariance = target_covariance - right_term
  
  new_covariance = as.matrix(Matrix::nearPD(new_covariance)$mat)
  
  
  # print("check positive definite, new covaraince")
  # print(matrixcalc::is.positive.definite(new_covariance))
  
  
  randomized_y = matrix(rep(0, nrow(unprotected_data) * length(target_indices)), nrow = nrow(unprotected_data), byrow = TRUE)
  
  for( index in 1:nrow(unprotected_data))
  {
    new_means = target_means + new_y_variates_covariance %*%
      solve(unprotected_cov) %*% (unprotected_data[index,]- unprotected_means)
    
    new_means = t(new_means)
    
    new_rand_vars = MASS::mvrnorm(n = 1, mu = new_means, Sigma =  new_covariance)
    
    
    
    randomized_y[index,] = new_rand_vars
  }
  
  # print(new_covariance)
  
  colnames(randomized_y) = paste0("y_", 1:length(target_indices))
  
  return(randomized_y)
  
}

gadp_func_ver_2 = function(unprotected_data, target_indices, theta = 0.4, dat.seed = 12)
{
  set.seed(dat.seed)
  
  unprotected_data = as.matrix(unprotected_data)
  dat_means = colMeans(unprotected_data)
  # means of the target variable
  target_means = dat_means[target_indices]
  # means of the auxiliary variable
  auxiliary_means = dat_means[-target_indices]
  # Covariance of the target variable
  target_covariance = cov(unprotected_data)[target_indices,target_indices]
  print("target covariance")
  print(target_covariance)
  
  # Auxiliary variables covariance
  aux_covariance = cov(unprotected_data[,-target_indices])
  
  target_aux_covariance = as.matrix(Matrix::nearPD(cov(unprotected_data))$mat)
  target_aux_covariance = 
    target_aux_covariance[-c(target_indices),
                          c(target_indices)]
  
  # print("auxiliary covariance matrix")
  # print(dim(target_aux_covariance))
  # 
  # print("target covariance matrix")
  # print(dim(target_covariance))
  
  unprotected_cov = cov(unprotected_data)
  
  unprotected_means = colMeans(unprotected_data)
  
  
  
  # random_variates = MASS::mvrnorm(mu = c(target_means, auxiliary_means),
  #                                 Sigma = )
  
  new_y_variates_covariance = unprotected_cov[target_indices,target_indices] * theta^2
  if(length(new_y_variates_covariance) == 1)
  {
    new_y_variates_covariance = cbind(t(unprotected_cov[target_indices, -target_indices]), 
                                      c(new_y_variates_covariance))
  }
  else{
    new_y_variates_covariance = cbind(unprotected_cov[target_indices, -target_indices], 
                                      new_y_variates_covariance)
  }
  
  
  # print("Randomized y covariance dimensions")
  # print(dim(new_y_variates_covariance))
  # 
  # print("target covariance")
  # print(target_covariance)
  # print(dim(target_covariance))
  # print("covariance for new variables")
  # print(new_y_variates_covariance)
  # print("covarinace for unprotected variables")
  # print(unprotected_cov)
  
  print(new_y_variates_covariance)
  print(dim(new_y_variates_covariance))
  print(dim(solve(unprotected_cov)))
  
  right_term = new_y_variates_covariance %*%
    solve(unprotected_cov) %*% t(new_y_variates_covariance)
  
  print(right_term)
  
  right_term = as.matrix(Matrix::nearPD(right_term)$mat)
  
  # print(right_term)
  # 
  # print("check positive definite")
  #   print(matrixcalc::is.positive.definite(right_term))
  #   print(matrixcalc::is.positive.definite(target_covariance))
  
  
  new_covariance = target_covariance - right_term
  print(new_covariance)
  new_covariance = as.matrix(Matrix::nearPD(new_covariance)$mat)
  print(new_covariance)
  
  # print("check positive definite, new covaraince")
  # print(matrixcalc::is.positive.definite(new_covariance))
  
  
  randomized_y = matrix(rep(0, nrow(unprotected_data) * length(target_indices)), nrow = nrow(unprotected_data), byrow = TRUE)
  
  for( index in 1:nrow(unprotected_data))
  {
    new_means = target_means + new_y_variates_covariance %*%
      solve(unprotected_cov) %*% (unprotected_data[index,]- unprotected_means)
    
    new_means = t(new_means)
    
    new_rand_vars = MASS::mvrnorm(n = 1, mu = new_means, Sigma =  new_covariance)
    
    
    
    randomized_y[index,] = new_rand_vars
  }
  
  # print(new_covariance)
  
  colnames(randomized_y) = paste0("y_", 1:length(target_indices))
  
  return(randomized_y)
  
}

# Model 1 CGADP ----------------
# Can work for Model 3 as well as it is standard uniform and  the Y's 
# appear normally distributed
model_1_cgadp = function(unprotected_data, target_indices, 
                         theta = 0.4, dat.seed = 12)
{
  
  # Set seed for generating random multivariate normal units
  set.seed(dat.seed)
  
  unprotected_data = as.matrix(unprotected_data)
  
  y_names = colnames(unprotected_data[,target_indices])
  
  
  
  
  # To be released as it is not sensitive.
  original_x = unprotected_data[,-target_indices]
  
  # convert to standard uniform (for sanity check)
  unprotected_data_x = apply(unprotected_data[,-target_indices],2,punif)
  # Convert to standard normal distibution
  unprotected_data_x = apply(unprotected_data_x, 2, qnorm, mean = 0, sd = 1)
  
  
  unprotected_data_y = unprotected_data[,target_indices]
  
  y_means = numeric(length(target_indices))
  
  y_sds = numeric(length(target_indices))
  
  # y looks to be closely normally distributed
  for(y_index in 1:length(target_indices))
  {
    # Calculate the empirical mean and standard deviation
    y_means[y_index] = mean(unprotected_data_y[,y_index])
    y_sds[y_index] = sd(unprotected_data_y[,y_index])
    # Use cumulative function to convert to standard uniform
    y_cumulative = pnorm(unprotected_data_y[,y_index], y_means[y_index],
                         y_sds[y_index])
    # Covert to standard normal
    y_stan_normal = qnorm(y_cumulative, mean = 0, sd = 1)
    unprotected_data_y[,y_index] = y_stan_normal
    
  }
  # Initial preprocessing finished
  unprotected_data = cbind(unprotected_data_x, unprotected_data_y)
  
  # Spearman rank correlation matrix
  spearman_rank_corr = cor(unprotected_data, method = "spearman")
  
  # Claculate the Pearson correlation matrix
  pearson_corr = matrix(rep(0, nrow(spearman_rank_corr) * nrow(spearman_rank_corr)),
                        nrow = nrow(spearman_rank_corr), byrow = TRUE)
  # Derive Pearson correlation matrix from Spearman rank correlation matrix
  for(row_index_i in 1:ncol(unprotected_data))
  {
    for(col_index_j in 1:ncol(unprotected_data))
    {
      
      pearson_corr[row_index_i, col_index_j] = 
        2 * sin((pi * spearman_rank_corr[row_index_i, col_index_j]) / 6)
    }
  }
  
  
  #print("Constructed Pearson correlation matrix")
  
  # https://blogs.sas.com/content/iml/2010/12/10/converting-between-correlation-and-covariance-matrices.html
  # By Rick Wicklin on DoLoop
  # https://stats.stackexchange.com/questions/62850/obtaining-covariance-matrix-from-correlation-matrix
  # Community Wiki
  #pearson_corr = as.matrix(Matrix::nearPD(pearson_corr)$mat)
  
  # Standard deviation should be one as the distributions are standard normal.
  #stan_dev_matrix = diag(ncol(unprotected_data))
  
  #covariance_mat = stan_dev_matrix %*% pearson_corr %*% stan_dev_matrix
  
  covariance_mat = as.matrix(nearPD(pearson_corr)$mat)
  
  # means of the target variable
  target_means = colMeans(unprotected_data[,target_indices])
  
  # means of the auxiliary variable
  auxiliary_means = colMeans(unprotected_data[,-target_indices])
  # Covariance of the target variable
  target_covariance = covariance_mat[target_indices,target_indices]
  
  # Covariance calculated 
  # Auxiliary variables covariance
  aux_covariance = covariance_mat[,-target_indices]
  
  target_aux_covariance = covariance_mat[-target_indices,target_indices]
  
  

  # update the unprotected data covariance
  unprotected_cov = covariance_mat
  # Get the unprotected data column means
  unprotected_means = colMeans(unprotected_data)
  
  # Using the new Pearson from the Spearman covariance matrix
  new_y_variates_covariance = unprotected_cov[target_indices,target_indices] * theta^2
  new_y_variates_covariance = cbind(unprotected_cov[target_indices,-target_indices] , 
                                    new_y_variates_covariance)
  
  right_term = new_y_variates_covariance %*%
    solve(unprotected_cov, tol=NULL) %*% t(new_y_variates_covariance)
  
  new_covariance = target_covariance - right_term
  
  # Make sure covariance matrixis near positive definite
  new_covariance = as.matrix(nearPD(new_covariance, eig.tol = 1e-10, 
                                    maxit = 200, posd.tol = 1e-09,
                                    conv.tol = 1e-09)$mat)
  
  # For storing the privatized y, or y after applying gadp
  randomized_y = matrix(rep(0, nrow(unprotected_data) * length(target_indices)), nrow = nrow(unprotected_data), byrow = TRUE)
  
  for( index in 1:nrow(unprotected_data))
  {
    new_means = target_means + new_y_variates_covariance %*%
      solve(unprotected_cov) %*% (unprotected_data[index,]- unprotected_means)
    
    new_means = t(new_means)
    
    new_rand_vars = MASS::mvrnorm(n = 1, mu = new_means, Sigma =  new_covariance)
    
    
    
    randomized_y[index,] = new_rand_vars
  }
  
  
  # Convert the y-values to uniform distribution to convert back to 
  for(y_index in 1:length(target_indices))
  {
    # Convert the distributions back to the uniform distribution
    new_randomized_mean = mean(randomized_y[,y_index])
    new_randomized_sd = sd(randomized_y[,y_index])
    randomized_y_unif = pnorm(randomized_y[,y_index], mean = new_randomized_mean,
                                   sd = new_randomized_sd)
    
    # Prevent infinities, use approximations ot 0 and 1,
    randomized_y_unif = scales::rescale(randomized_y_unif, to = c(0 + .Machine$double.eps, 1 - .Machine$double.eps))
    # Convert the y-values back to their original distribution
    
    y_index_original_dist = qnorm(randomized_y_unif,
                                  mean = y_means[y_index], sd = y_sds[y_index])
    # Update the randomized values to no longer be stnadard normal, but the in the
    # same distirbution as the original y-values, but not the original y-values
    # themselves
    # print("sum of infinites")
    # print(sum(is.infinite(y_index_original_dist)))
    # print(y_means[y_index]) 
    # print(y_sds[y_index])
    
    randomized_y[,y_index] = y_index_original_dist
  }
  
  colnames(randomized_y) = y_names
  
  randomized_data = cbind(original_x, randomized_y)
  
  return(randomized_data)
}

## Model 1 CGADP Version 2 ----------------
# Underwent code refactoring to try and improve perfromance
model_1_cgadp_ver_2 = function(unprotected_data, target_indices, 
                         theta = 0.4, dat.seed = 12)
{
  # Produces same output as version 1
  
  # Set seed for generating random multivariate normal units
  set.seed(dat.seed)
  
  unprotected_data = as.matrix(unprotected_data)
  
  y_names = colnames(unprotected_data[,target_indices])
  
  
  
  
  # To be released as it is not sensitive.
  original_x = unprotected_data[,-target_indices]
  
  # convert to standard uniform (for sanity check)
  unprotected_data_x = apply(unprotected_data[,-target_indices],2,punif)
  # Convert to standard normal distibution
  unprotected_data_x = apply(unprotected_data_x, 2, qnorm, mean = 0, sd = 1)
  
  
  unprotected_data_y = unprotected_data[,target_indices]
  
  y_means = numeric(length(target_indices))
  
  y_sds = numeric(length(target_indices))
  
  # y looks to be closely normally distributed
  for(y_index in 1:length(target_indices))
  {
    # Calculate the empirical mean and standard deviation
    y_means[y_index] = mean(unprotected_data_y[,y_index])
    y_sds[y_index] = sd(unprotected_data_y[,y_index])
    # Use cumulative function to convert to standard uniform
    y_cumulative = pnorm(unprotected_data_y[,y_index], y_means[y_index],
                         y_sds[y_index])
    # Covert to standard normal
    y_stan_normal = qnorm(y_cumulative, mean = 0, sd = 1)
    unprotected_data_y[,y_index] = y_stan_normal
    
  }
  # Initial preprocessing finished
  unprotected_data = cbind(unprotected_data_x, unprotected_data_y)
  
  # Spearman rank correlation matrix
  spearman_rank_corr = cor(unprotected_data, method = "spearman")
  
  # Claculate the Pearson correlation matrix
  pearson_corr = matrix(rep(0, nrow(spearman_rank_corr) * nrow(spearman_rank_corr)),
                        nrow = nrow(spearman_rank_corr), byrow = TRUE)
  # Derive Pearson correlation matrix from Spearman rank correlation matrix
  for(row_index_i in 1:ncol(unprotected_data))
  {
    for(col_index_j in 1:ncol(unprotected_data))
    {
      
      pearson_corr[row_index_i, col_index_j] = 
        2 * sin((pi * spearman_rank_corr[row_index_i, col_index_j]) / 6)
    }
  }
  
  
  #print("Constructed Pearson correlation matrix")
  
  # https://blogs.sas.com/content/iml/2010/12/10/converting-between-correlation-and-covariance-matrices.html
  # By Rick Wicklin on DoLoop
  # https://stats.stackexchange.com/questions/62850/obtaining-covariance-matrix-from-correlation-matrix
  # Community Wiki
  #pearson_corr = as.matrix(Matrix::nearPD(pearson_corr)$mat)
  
  # Standard deviation should be one as the distributions are standard normal.
  #stan_dev_matrix = diag(ncol(unprotected_data))
  
  #covariance_mat = stan_dev_matrix %*% pearson_corr %*% stan_dev_matrix
  
  covariance_mat = as.matrix(nearPD(pearson_corr)$mat)
  
  # means of the target variable
  target_means = colMeans(unprotected_data[,target_indices])
  
  # means of the auxiliary variable
  auxiliary_means = colMeans(unprotected_data[,-target_indices])
  # Covariance of the target variable
  target_covariance = covariance_mat[target_indices,target_indices]
  
  # Covariance calculated 
  # Auxiliary variables covariance
  aux_covariance = covariance_mat[,-target_indices]
  
  target_aux_covariance = covariance_mat[-target_indices,target_indices]
  
  
  
  # update the unprotected data covariance
  unprotected_cov = covariance_mat
  # Get the unprotected data column means
  unprotected_means = colMeans(unprotected_data)
  
  # Using the new Pearson from the Spearman covariance matrix
  new_y_variates_covariance = unprotected_cov[target_indices,target_indices] * theta^2
  new_y_variates_covariance = cbind(unprotected_cov[target_indices,-target_indices] , 
                                    new_y_variates_covariance)
  
  right_term = new_y_variates_covariance %*%
    solve(unprotected_cov, tol=NULL) %*% t(new_y_variates_covariance)
  
  new_covariance = target_covariance - right_term
  
  # Make sure covariance matrixis near positive definite
  new_covariance = as.matrix(nearPD(new_covariance, eig.tol = 1e-10, 
                                    maxit = 200, posd.tol = 1e-09,
                                    conv.tol = 1e-09)$mat)
  
  # For storing the privatized y, or y after applying gadp
  randomized_y = matrix(rep(0, nrow(unprotected_data) * length(target_indices)), nrow = nrow(unprotected_data), byrow = TRUE)
  
  multi_var_norm_deviates = function(x_row, 
                                     target_mus = target_means, 
                                     new_covariates_sigma = new_y_variates_covariance, 
                                     unprotected_cov_mat = unprotected_cov,
                                     new_sigma = new_covariance, 
                                     unprotected_mus = unprotected_means)
  {
    new_means = target_mus + new_covariates_sigma %*%
      solve(unprotected_cov_mat) %*% (x_row - unprotected_means)
    
    new_means = t(new_means)
    
    new_rand_vars = MASS::mvrnorm(n = 1, mu = new_means, Sigma =  new_sigma)
    
    
    return(new_rand_vars)
  }
  
  randomized_y = t(apply(unprotected_data, MARGIN = 1, FUN = multi_var_norm_deviates))
  
  
  # Convert the y-values to uniform distribution to convert back to 
  for(y_index in 1:length(target_indices))
  {
    # Convert the distributions back to the uniform distribution
    new_randomized_mean = mean(randomized_y[,y_index])
    new_randomized_sd = sd(randomized_y[,y_index])
    randomized_y_unif = pnorm(randomized_y[,y_index], mean = new_randomized_mean,
                              sd = new_randomized_sd)
    
    # Prevent infinities, use approximations ot 0 and 1,
    randomized_y_unif = scales::rescale(randomized_y_unif, to = c(0 + .Machine$double.eps, 1 - .Machine$double.eps))
    # Convert the y-values back to their original distribution
    
    y_index_original_dist = qnorm(randomized_y_unif,
                                  mean = y_means[y_index], sd = y_sds[y_index])
    # Update the randomized values to no longer be stnadard normal, but the in the
    # same distirbution as the original y-values, but not the original y-values
    # themselves
    # print("sum of infinites")
    # print(sum(is.infinite(y_index_original_dist)))
    # print(y_means[y_index]) 
    # print(y_sds[y_index])
    
    randomized_y[,y_index] = y_index_original_dist
  }
  
  colnames(randomized_y) = y_names
  
  randomized_data = cbind(original_x, randomized_y)
  
  return(randomized_data)
}

# Model 2 CGADP  -------------------
# Incorporates an EM algorithm to identify the means and standard deviations of 
# the two normal distributions in mixture distribution of Model 2
model_2_cgadp_ver_1 = function(unprotected_data, target_indices, 
                               theta = 0.4, dat.seed = 12)
{
  # Produces same output as version 1
  
  # Set seed for generating random multivariate normal units
  set.seed(dat.seed)
  
  unprotected_data = as.matrix(unprotected_data)
  
  y_names = colnames(unprotected_data[,target_indices])
  
  
  
  
  # To be released as it is not sensitive.
  original_x = unprotected_data[,-target_indices]
  
  # convert to standard uniform (for sanity check)
  unprotected_data_x = apply(unprotected_data[,-target_indices],2,pnorm)
  # Convert to standard normal distibution
  unprotected_data_x = apply(unprotected_data_x, 2, qnorm, mean = 0, sd = 1)
  
  
  unprotected_data_y = unprotected_data[,target_indices]
  
  # Need a martrix for the two means of the mixture
  y_means = matrix(rep(0, length(target_indices) * 2), nrow = 2, byrow = TRUE)
  # Need a martrix for the two standard deviations of the mixture
  y_sds = matrix(rep(0, length(target_indices) * 2), nrow = 2, byrow = TRUE)
  
  mixed_norm_est_list = list()
  # y looks to be closely normally distributed
  for(y_index in 1:length(target_indices))
  {
    # Use EM algorithm to find the means and standard deviations of the mixture
    # of two normal distributions for each y
    mixed_norm_est_list[[y_index]] = 
      mixtools::normalmixEM(unprotected_data_y[,y_index], lambda = 0.5, k = 2)
    
    mixed_norm_estimates = mixed_norm_est_list[[y_index]]
    # Calculate the empirical mean and standard deviation
    # then store in matrices
    y_means[y_index,] = mixed_norm_estimates$mu
    
    
    y_sds[y_index,] = mixed_norm_estimates$sigma
    # Use cumulative function to convert to standard uniform
    y_cumulative = EnvStats::pnormMix(unprotected_data_y[,y_index], 
                                      mean1 = y_means[y_index,1],
                                      mean2 = y_means[y_index,2], 
                                      sd1 = y_sds[y_index,1],
                                      sd2 = y_sds[y_index,2], 
                                      p.mix = mixed_norm_estimates$lambda)
    
    # Covert to standard normal
    y_stan_normal = qnorm(y_cumulative, mean = 1, sd = 1)
    unprotected_data_y[,y_index] = y_stan_normal
    
  }
  
  # Initial preprocessing finished
  unprotected_data = cbind(unprotected_data_x, unprotected_data_y)
  
  # Spearman rank correlation matrix
  spearman_rank_corr = cor(unprotected_data, method = "spearman")
  
  # Calculate the Pearson correlation matrix
  # First construct the eampty pearson correlation matrix
  pearson_corr = matrix(rep(0, nrow(spearman_rank_corr) * nrow(spearman_rank_corr)),
                        nrow = nrow(spearman_rank_corr), byrow = TRUE)
  # Calculate Pearson correlation matrix from Spearman rank correlation matrix
  pearson_corr = 
    2 * sin((pi * spearman_rank_corr) / 6)
  
  
  
  # https://blogs.sas.com/content/iml/2010/12/10/converting-between-correlation-and-covariance-matrices.html
  # By Rick Wicklin on DoLoop
  # https://stats.stackexchange.com/questions/62850/obtaining-covariance-matrix-from-correlation-matrix
  # Community Wiki
  #pearson_corr = as.matrix(Matrix::nearPD(pearson_corr)$mat)
  
  # Standard deviation should be one as the distributions are standard normal.
  #stan_dev_matrix = diag(ncol(unprotected_data))
  
  #covariance_mat = stan_dev_matrix %*% pearson_corr %*% stan_dev_matrix
  
  covariance_mat = as.matrix(nearPD(pearson_corr)$mat)
  
  # means of the target variable
  target_means = colMeans(unprotected_data[,target_indices])
  
  # means of the auxiliary variable
  auxiliary_means = colMeans(unprotected_data[,-target_indices])
  # Covariance of the target variable
  target_covariance = covariance_mat[target_indices,target_indices]
  
  # Covariance calculated 
  # Auxiliary variables covariance
  aux_covariance = covariance_mat[,-target_indices]
  # Covariance for the response variables
  target_aux_covariance = covariance_mat[-target_indices,target_indices]
  
  
  
  # update the unprotected data covariance
  unprotected_cov = covariance_mat
  # Get the unprotected data column means
  unprotected_means = colMeans(unprotected_data)
  
  # Using the new Pearson from the Spearman covariance matrix
  new_y_variates_covariance = unprotected_cov[target_indices,target_indices] * theta^2
  new_y_variates_covariance = cbind(unprotected_cov[target_indices,-target_indices] , 
                                    new_y_variates_covariance)
  # Far right term to adjut the variance to of the new random normal deviates
  # focusing on the sensitive response values
  right_term = new_y_variates_covariance %*%
    solve(unprotected_cov, tol=NULL) %*% t(new_y_variates_covariance)
  
  new_covariance = target_covariance - right_term
  
  # Make sure covariance matrixis near positive definite
  new_covariance = as.matrix(nearPD(new_covariance, eig.tol = 1e-10, 
                                    maxit = 200, posd.tol = 1e-09,
                                    conv.tol = 1e-09)$mat)
  
  
  # For storing the privatized y, or y after applying gadp
  randomized_y = matrix(rep(0, nrow(unprotected_data) * length(target_indices)), nrow = nrow(unprotected_data), byrow = TRUE)
  
  # function for generating the random normal deviates.
  multi_var_norm_deviates = function(x_row, 
                                     target_mus = target_means, 
                                     new_covariates_sigma = new_y_variates_covariance, 
                                     unprotected_cov_mat = unprotected_cov,
                                     new_sigma = new_covariance, 
                                     unprotected_mus = unprotected_means)
  {
    # Calculate the new means per element
    new_means = target_mus + new_covariates_sigma %*%
      solve(unprotected_cov_mat) %*% (x_row - unprotected_means)
    # Transpose to accomodate mvrnorm
    new_means = t(new_means)
    # generate new random deviates
    new_rand_vars = MASS::mvrnorm(n = 1, mu = new_means, Sigma =  new_sigma)
    
    
    return(new_rand_vars)
  }
  # Transpose the new response values to combine with x-matrix later
  randomized_y = t(apply(unprotected_data, MARGIN = 1, FUN = multi_var_norm_deviates))
  
  # Convert the y-values to uniform distribution to convert back to 
  for(y_index in 1:length(target_indices))
  {
    # Convert the distributions back to the uniform distribution
    new_randomized_mean = mean(randomized_y[,y_index])
    new_randomized_sd = sd(randomized_y[,y_index])
    # Standard uniform distribution
    randomized_y_unif = pnorm(randomized_y[,y_index], mean = new_randomized_mean,
                              sd = new_randomized_sd)
    
    # Prevent infinities, use approximations to 0 and 1,
    randomized_y_unif = scales::rescale(randomized_y_unif, to = c(0 + .Machine$double.eps, 1 - .Machine$double.eps))
    
    # Convert the y-values back to their original distribution
    mixed_norm_estimates = mixed_norm_est_list[[y_index]]
    # Getting the mixture of two normal distributions
    y_index_original_dist = EnvStats::qnormMix(randomized_y_unif, mean1 = y_means[y_index,1],
                                               mean2 = y_means[y_index,2], 
                                               sd1 = y_sds[y_index,1],
                                               sd2 = y_sds[y_index,2], 
                                               p.mix = mixed_norm_estimates$lambda[2])
    # Update the randomized values to no longer be standard normal, but the in the
    # same distribution as the original y-values, but not the original y-values
    # themselves
    
    randomized_y[,y_index] = y_index_original_dist
  }
  # put the names for the response back
  colnames(randomized_y) = y_names
  
  randomized_data = cbind(original_x, randomized_y)
  
  return(randomized_data)
}

# CGADP for CE data -----------------------
# Adapted from Model 2 CGADP code
real_CE_cgadp_ver_1 = function(unprotected_data, target_indices, 
                               theta = 0.4, dat.seed = 12)
{
  # Highly specific due to the nature of the CGADP algorithm
  # Adding or removing variables needs to happen manually
  transform_CE_data = function(CE_data)
  {
    # Using maximum likelihood to estimate the parameters of each distribution
    # Each variable first has the ML estimates of the parameters calculated,
    # then the data is updated using the PDF. The result is every variable
    # is converted to uniform distribution
    # params_EHOUSNGC = EnvStats::eexp(CE_data[,"EHOUSNGC"])
    # 
    # CE_data[,"EHOUSNGC"] = pexp(CE_data[,"EHOUSNGC"], 
    #                             rate = params_EHOUSNGC$parameters)
    
    # 
    # params_IRAX = EnvStats::eexp(CE_data[,"IRAX"])
    
    # CE_data[,"IRAX"] = pexp( CE_data[,"IRAX"], rate = params_IRAX$parameters)
    # 
    # params_LIQUIDX = EnvStats::eexp(CE_data[,"LIQUIDX"])
    # 
    # CE_data[,"LIQUIDX"] = pexp( CE_data[,"LIQUIDX"], rate = params_LIQUIDX$parameters)
    # 
    params_SEX = EnvStats::ebinom(CE_data[,"SEXM"])
    
    CE_data[,"SEXM"] = pbinom(CE_data[,"SEXM"], 
                              1, params_SEX$parameters[2])
    
    print(colnames(CE_data))
    params_AGE = mixtools::normalmixEM(CE_data[,"AGE"], lambda = 0.5, k = 2)
    
    print( params_AGE$lambda)
    
    
    CE_data[,"AGE"] = EnvStats::pnormMix(CE_data[,"AGE"], 
                                         mean1 = params_AGE$mu[1],
                                         mean2 = params_AGE$mu[2], 
                                         sd1 = params_AGE$sigma[1],
                                         sd2 = params_AGE$sigma[2], 
                                         p.mix = params_AGE$lambda)
    
    # params_STOCKX = EnvStats::eexp(CE_data[,"STOCKX"])
    # 
    # CE_data[,"STOCKX"] = pexp(CE_data[,"STOCKX"], rate = params_STOCKX$parameters)
    
    # params_STUDNTX = EnvStats::eexp(CE_data[,"STUDNTX"])
    # 
    # CE_data[,"STUDNTX"] = pexp(CE_data[,"STUDNTX"], rate = params_STUDNTX$parameters)
    # Note have to shift distribution of TOTXEST as there are negative values present
    print("run 3")
    params_TOTXEST = EnvStats::elnorm(CE_data[,"TOTXEST"])
    
    CE_data[,"TOTXEST"] = plnorm(CE_data[,"TOTXEST"], 
                                 meanlog = params_TOTXEST$parameters[1],
                                 sdlog = params_TOTXEST$parameters[2])
    
    params_FINCBTAX = EnvStats::elnorm(CE_data[,"y_1"])
    
    CE_data[,"y_1"] = plnorm(CE_data[,"y_1"], 
                             meanlog = params_FINCBTAX$parameters[1],
                             sdlog = params_FINCBTAX$parameters[2])
    
    params_FINCBTAX2 = EnvStats::elnorm(CE_data[,"y_2"])
    
    CE_data[,"y_2"] = plnorm(CE_data[,"y_2"], 
                             meanlog = params_FINCBTAX2$parameters[1],
                             sdlog = params_FINCBTAX2$parameters[2])
    
    params_TOTEXPCQ = EnvStats::elnorm(CE_data[,"TOTEXPCQ"])
    
    CE_data[,"TOTEXPCQ"] = plnorm(CE_data[,"TOTEXPCQ"], 
                                  meanlog = params_TOTEXPCQ$parameters[1],
                                  sdlog = params_TOTEXPCQ$parameters[2])
    
    print("run 4")
    # FInal object with ML estimates information as well as the transformed 
    # data
    transformed_data_info = 
      list("AGE" = params_AGE,
           "SEX" = params_SEX,
           "TOTEXPCQ" = params_TOTEXPCQ,
           "TOTXEST" = params_TOTXEST,
           "FINCBTAX" = params_FINCBTAX,
           "FINCBTAX2" = params_FINCBTAX2,
           "dat" = CE_data)
    
    return(transformed_data_info)
  }
  
  # Used to undo transformations on the response in particular
  undo_transform_CE_data = function(calculated_params,CE_data_resp)
  {
    
    # Using previously calculated ML estimates of the distributions to be able to 
    # transform the reponses back to their original distirbution
    CE_data_resp[,"y_1"] = qlnorm(CE_data_resp[,"y_1"],
                                  meanlog = calculated_params$FINCBTAX$parameters[1],
                                  sdlog = calculated_params$FINCBTAX$parameters[2])
    
    CE_data_resp[,"y_2"] = qlnorm(CE_data_resp[,"y_2"],
                                  meanlog = calculated_params$FINCBTAX2$parameters[1],
                                  sdlog = calculated_params$FINCBTAX2$parameters[2])
    
    # Transformed back to original distribution
    CE_data_resp
    
    return(CE_data_resp)
  }
  
  # Set seed for generating random multivariate normal units
  set.seed(dat.seed)
  
  unprotected_data = as.matrix(unprotected_data)
  
  y_names = colnames(unprotected_data[,target_indices])
  
  
  
  
  # To be released as it is not sensitive.
  original_x = unprotected_data[,-target_indices]

  
  uniform_dist_data_info = transform_CE_data(unprotected_data)
  
  # Initial preprocessing finished
  unprotected_data = uniform_dist_data_info$dat
  
  # Rescaling to avoid infinities
  unprotected_data_1 = apply(unprotected_data, 2, 
                           FUN = function(x){
                             shifted_x = scales::rescale(x,c(0+ .Machine$double.eps, 1 - .Machine$double.eps))
                           return(shifted_x)})
  # Data augmentation to ensure binary variable does not generate infinities
  unprotected_data_1[,"SEXM"] = ifelse(unprotected_data[,"SEXM"] == 1, 1- .Machine$double.eps,
                                       unprotected_data[,"SEXM"])
  
  unprotected_data = apply(unprotected_data_1, 2, qnorm, mean = 0, sd = 1)
  
  
  # Spearman rank correlation matrix
  spearman_rank_corr = cor(unprotected_data, method = "spearman")
  
  # Calculate the Pearson correlation matrix
  # First construct the eampty pearson correlation matrix
  pearson_corr = matrix(rep(0, nrow(spearman_rank_corr) * nrow(spearman_rank_corr)),
                        nrow = nrow(spearman_rank_corr), byrow = TRUE)
  # Calculate Pearson correlation matrix from Spearman rank correlation matrix
  pearson_corr = 
    2 * sin((pi * spearman_rank_corr) / 6)
  
  # https://blogs.sas.com/content/iml/2010/12/10/converting-between-correlation-and-covariance-matrices.html
  # By Rick Wicklin on DoLoop
  # https://stats.stackexchange.com/questions/62850/obtaining-covariance-matrix-from-correlation-matrix
  # Community Wiki
  #pearson_corr = as.matrix(Matrix::nearPD(pearson_corr)$mat)
  
  # Standard deviation should be one as the distributions are standard normal.
  #stan_dev_matrix = diag(ncol(unprotected_data))
  
  #covariance_mat = stan_dev_matrix %*% pearson_corr %*% stan_dev_matrix
  
  covariance_mat = as.matrix(nearPD(pearson_corr)$mat)
  
  # means of the target variable
  target_means = colMeans(unprotected_data[,target_indices])
  
  # means of the auxiliary variable
  auxiliary_means = colMeans(unprotected_data[,-target_indices])
  # Covariance of the target variable
  target_covariance = covariance_mat[target_indices,target_indices]
  
  # Covariance calculated 
  # Auxiliary variables covariance
  aux_covariance = covariance_mat[,-target_indices]
  # Covariance for the response variables
  target_aux_covariance = covariance_mat[-target_indices,target_indices]
  
  
  
  # update the unprotected data covariance
  unprotected_cov = covariance_mat
  # Get the unprotected data column means
  unprotected_means = colMeans(unprotected_data)
  
  # Using the new Pearson from the Spearman covariance matrix
  new_y_variates_covariance = unprotected_cov[target_indices,target_indices] * theta^2
  new_y_variates_covariance = cbind(unprotected_cov[target_indices,-target_indices] , 
                                    new_y_variates_covariance)
  # Far right term to adjut the variance to of the new random normal deviates
  # focusing on the sensitive response values
  right_term = new_y_variates_covariance %*%
    solve(unprotected_cov, tol=NULL) %*% t(new_y_variates_covariance)
  
  new_covariance = target_covariance - right_term
  
  # Make sure covariance matrixis near positive definite
  new_covariance = as.matrix(nearPD(new_covariance, eig.tol = 1e-10, 
                                    maxit = 200, posd.tol = 1e-09,
                                    conv.tol = 1e-09)$mat)
  
  
  # For storing the privatized y, or y after applying gadp
  randomized_y = matrix(rep(0, nrow(unprotected_data) * length(target_indices)), nrow = nrow(unprotected_data), byrow = TRUE)
  
  print(head(unprotected_data))
  
  # function for generating the random normal deviates.
  multi_var_norm_deviates = function(x_row, 
                                     target_mus = target_means, 
                                     new_covariates_sigma = new_y_variates_covariance, 
                                     unprotected_cov_mat = unprotected_cov,
                                     new_sigma = new_covariance, 
                                     unprotected_mus = unprotected_means)
  {
    # Calculate the new means per element
    new_means = target_mus + new_covariates_sigma %*%
      solve(unprotected_cov_mat) %*% (x_row - unprotected_means)
    # Transpose to accomodate mvrnorm
    new_means = t(new_means)
    # generate new random deviates
    new_rand_vars = MASS::mvrnorm(n = 1, mu = new_means, Sigma =  new_sigma)
    
    
    return(new_rand_vars)
  }
  # Transpose the new response values to combine with x-matrix later
  randomized_y = t(apply(unprotected_data, MARGIN = 1, FUN = multi_var_norm_deviates))
  
  # Transform back to uniform distribution
  # Otherwise you get errors because you assume standard normal distirbuion is uniform distribution
  for(y_index in 1:length(target_indices))
  {
    # Convert the distributions back to the uniform distribution
    new_randomized_mean = mean(randomized_y[,y_index])
    new_randomized_sd = sd(randomized_y[,y_index])
    randomized_y_unif = pnorm(randomized_y[,y_index], mean = new_randomized_mean,
                              sd = new_randomized_sd)
    
    # Prevent infinities, use approximations ot 0 and 1,
    randomized_y_unif = scales::rescale(randomized_y_unif, to = c(0 + .Machine$double.eps, 1 - .Machine$double.eps))
    # Convert the y-values back to their original distribution
    # Will convert from uniform distirbution back to original distribution in next step
    randomized_y[,y_index] = randomized_y_unif
  }
  
  # Convert the y-values to uniform distribution to convert back to 
  randomized_y = undo_transform_CE_data(uniform_dist_data_info, randomized_y)
  
  
  # put the names for the response back
  colnames(randomized_y) = y_names
  
  randomized_data = cbind(original_x, randomized_y)
  
  return(randomized_data)
}



# Adapted from Model 2 CGADP code
# Uses SALARYX as well as FINCBTAX
real_CE_cgadp_ver_2 = function(unprotected_data, target_indices, 
                               theta = 0.4, dat.seed = 12)
{
  # Highly specific due to the nature of the CGADP algorithm
  # Adding or removing variables needs to happen manually
  transform_CE_data = function(CE_data)
  {
    # Using maximum likelihood to estimate the parameters of each distribution
    # Each variable first has the ML estimates of the parameters calculated,
    # then the data is updated using the PDF. The result is every variable
    # is converted to uniform distribution
    # params_EHOUSNGC = EnvStats::eexp(CE_data[,"EHOUSNGC"])
    # 
    # CE_data[,"EHOUSNGC"] = pexp(CE_data[,"EHOUSNGC"], 
    #                             rate = params_EHOUSNGC$parameters)
    
    # 
    # params_IRAX = EnvStats::eexp(CE_data[,"IRAX"])
    
    # CE_data[,"IRAX"] = pexp( CE_data[,"IRAX"], rate = params_IRAX$parameters)
    # 
    # params_LIQUIDX = EnvStats::eexp(CE_data[,"LIQUIDX"])
    # 
    # CE_data[,"LIQUIDX"] = pexp( CE_data[,"LIQUIDX"], rate = params_LIQUIDX$parameters)
    # 
    params_SEX = EnvStats::ebinom(CE_data[,"SEXM"])
    
    CE_data[,"SEXM"] = pbinom(CE_data[,"SEXM"], 
                              1, params_SEX$parameters[2])
    
    print(colnames(CE_data))
    params_AGE = mixtools::normalmixEM(CE_data[,"AGE"], lambda = 0.5, k = 2)
    
    print( params_AGE$lambda)
    
    
    CE_data[,"AGE"] = EnvStats::pnormMix(CE_data[,"AGE"], 
                                         mean1 = params_AGE$mu[1],
                                         mean2 = params_AGE$mu[2], 
                                         sd1 = params_AGE$sigma[1],
                                         sd2 = params_AGE$sigma[2], 
                                         p.mix = params_AGE$lambda)
    
    # params_STOCKX = EnvStats::eexp(CE_data[,"STOCKX"])
    # 
    # CE_data[,"STOCKX"] = pexp(CE_data[,"STOCKX"], rate = params_STOCKX$parameters)
    
    # params_STUDNTX = EnvStats::eexp(CE_data[,"STUDNTX"])
    # 
    # CE_data[,"STUDNTX"] = pexp(CE_data[,"STUDNTX"], rate = params_STUDNTX$parameters)
    # Note have to shift distribution of TOTXEST as there are negative values present
  
    params_TOTXEST = EnvStats::elnorm(CE_data[,"TOTXEST"])
    
    CE_data[,"TOTXEST"] = plnorm(CE_data[,"TOTXEST"], 
                                 meanlog = params_TOTXEST$parameters[1],
                                 sdlog = params_TOTXEST$parameters[2])
    
    params_FINCBTAX = EnvStats::elnorm(CE_data[,"y_1"])
    
    CE_data[,"y_1"] = plnorm(CE_data[,"y_1"], 
                             meanlog = params_FINCBTAX$parameters[1],
                             sdlog = params_FINCBTAX$parameters[2])
    
    params_SALARYX = MASS::fitdistr(CE_data[,"y_2"], densfun = "normal")
    
    CE_data[,"y_2"] = plnorm(CE_data[,"y_2"], 
                             mean = params_SALARYX$estimate[1],
                             sdlog = params_SALARYX$estimate[2])
    
    params_TOTEXPCQ = EnvStats::elnorm(CE_data[,"TOTEXPCQ"])
    
    CE_data[,"TOTEXPCQ"] = plnorm(CE_data[,"TOTEXPCQ"], 
                                  meanlog = params_TOTEXPCQ$parameters[1],
                                  sdlog = params_TOTEXPCQ$parameters[2])
    
    # FInal object with ML estimates information as well as the transformed 
    # data
    transformed_data_info = 
      list("AGE" = params_AGE,
           "SEX" = params_SEX,
           "TOTEXPCQ" = params_TOTEXPCQ,
           "TOTXEST" = params_TOTXEST,
           "FINCBTAX" = params_FINCBTAX,
           "SALARYX" = params_SALARYX,
           "dat" = CE_data)
    
    return(transformed_data_info)
  }
  
  # Used to undo transformations on the response in particular
  undo_transform_CE_data = function(calculated_params,CE_data_resp)
  {
    
    # Using previously calculated ML estimates of the distributions to be able to 
    # transform the reponses back to their original distirbution
    CE_data_resp[,"y_1"] = qlnorm(CE_data_resp[,"y_1"],
                                  meanlog = calculated_params$FINCBTAX$parameters[1],
                                  sdlog = calculated_params$FINCBTAX$parameters[2])
    
    CE_data_resp[,"y_2"] = qnorm(CE_data_resp[,"y_2"],
                                mean = calculated_params$SALARYX$estimate[1],
                                sd = calculated_params$SALARYX$estimate[2])
    
    # Transformed back to original distribution
    CE_data_resp
    
    return(CE_data_resp)
  }
  
  # Set seed for generating random multivariate normal units
  set.seed(dat.seed)
  
  unprotected_data = as.matrix(unprotected_data)
  
  y_names = colnames(unprotected_data[,target_indices])
  
  
  
  
  # To be released as it is not sensitive.
  original_x = unprotected_data[,-target_indices]
  
  
  uniform_dist_data_info = transform_CE_data(unprotected_data)
  
  # Initial preprocessing finished
  unprotected_data = uniform_dist_data_info$dat
  
  # Rescaling to avoid infinities
  unprotected_data_1 = apply(unprotected_data, 2, 
                             FUN = function(x){
                               shifted_x = scales::rescale(x,c(0+ .Machine$double.eps, 1 - .Machine$double.eps))
                               return(shifted_x)})
  # Data augmentation to ensure binary variable does not generate infinities
  unprotected_data_1[,"SEXM"] = ifelse(unprotected_data[,"SEXM"] == 1, 1- .Machine$double.eps,
                                       unprotected_data[,"SEXM"])
  
  unprotected_data = apply(unprotected_data_1, 2, qnorm, mean = 0, sd = 1)
  
  
  # Spearman rank correlation matrix
  spearman_rank_corr = cor(unprotected_data, method = "spearman")
  
  # Calculate the Pearson correlation matrix
  # First construct the eampty pearson correlation matrix
  pearson_corr = matrix(rep(0, nrow(spearman_rank_corr) * nrow(spearman_rank_corr)),
                        nrow = nrow(spearman_rank_corr), byrow = TRUE)
  # Calculate Pearson correlation matrix from Spearman rank correlation matrix
  pearson_corr = 
    2 * sin((pi * spearman_rank_corr) / 6)
  
  # https://blogs.sas.com/content/iml/2010/12/10/converting-between-correlation-and-covariance-matrices.html
  # By Rick Wicklin on DoLoop
  # https://stats.stackexchange.com/questions/62850/obtaining-covariance-matrix-from-correlation-matrix
  # Community Wiki
  #pearson_corr = as.matrix(Matrix::nearPD(pearson_corr)$mat)
  
  # Standard deviation should be one as the distributions are standard normal.
  #stan_dev_matrix = diag(ncol(unprotected_data))
  
  #covariance_mat = stan_dev_matrix %*% pearson_corr %*% stan_dev_matrix
  
  covariance_mat = as.matrix(nearPD(pearson_corr)$mat)
  
  # means of the target variable
  target_means = colMeans(unprotected_data[,target_indices])
  
  # means of the auxiliary variable
  auxiliary_means = colMeans(unprotected_data[,-target_indices])
  # Covariance of the target variable
  target_covariance = covariance_mat[target_indices,target_indices]
  
  # Covariance calculated 
  # Auxiliary variables covariance
  aux_covariance = covariance_mat[,-target_indices]
  # Covariance for the response variables
  target_aux_covariance = covariance_mat[-target_indices,target_indices]
  
  
  
  # update the unprotected data covariance
  unprotected_cov = covariance_mat
  # Get the unprotected data column means
  unprotected_means = colMeans(unprotected_data)
  
  # Using the new Pearson from the Spearman covariance matrix
  new_y_variates_covariance = unprotected_cov[target_indices,target_indices] * theta^2
  new_y_variates_covariance = cbind(unprotected_cov[target_indices,-target_indices] , 
                                    new_y_variates_covariance)
  # Far right term to adjut the variance to of the new random normal deviates
  # focusing on the sensitive response values
  right_term = new_y_variates_covariance %*%
    solve(unprotected_cov, tol=NULL) %*% t(new_y_variates_covariance)
  
  new_covariance = target_covariance - right_term
  
  # Make sure covariance matrixis near positive definite
  new_covariance = as.matrix(nearPD(new_covariance, eig.tol = 1e-10, 
                                    maxit = 200, posd.tol = 1e-09,
                                    conv.tol = 1e-09)$mat)
  
  
  # For storing the privatized y, or y after applying gadp
  randomized_y = matrix(rep(0, nrow(unprotected_data) * length(target_indices)), nrow = nrow(unprotected_data), byrow = TRUE)
  
  
  # function for generating the random normal deviates.
  multi_var_norm_deviates = function(x_row, 
                                     target_mus = target_means, 
                                     new_covariates_sigma = new_y_variates_covariance, 
                                     unprotected_cov_mat = unprotected_cov,
                                     new_sigma = new_covariance, 
                                     unprotected_mus = unprotected_means)
  {
    # Calculate the new means per element
    new_means = target_mus + new_covariates_sigma %*%
      solve(unprotected_cov_mat) %*% (x_row - unprotected_means)
    # Transpose to accomodate mvrnorm
    new_means = t(new_means)
    # generate new random deviates
    new_rand_vars = MASS::mvrnorm(n = 1, mu = new_means, Sigma =  new_sigma)
    
    
    return(new_rand_vars)
  }
  # Transpose the new response values to combine with x-matrix later
  randomized_y = t(apply(unprotected_data, MARGIN = 1, FUN = multi_var_norm_deviates))
  
  # Transform back to uniform distribution
  # Otherwise you get errors because you assume standard normal distirbuion is uniform distribution
  for(y_index in 1:length(target_indices))
  {
    # Convert the distributions back to the uniform distribution
    new_randomized_mean = mean(randomized_y[,y_index])
    new_randomized_sd = sd(randomized_y[,y_index])
    randomized_y_unif = pnorm(randomized_y[,y_index], mean = new_randomized_mean,
                              sd = new_randomized_sd)
    
    # Prevent infinities, use approximations ot 0 and 1,
    randomized_y_unif = scales::rescale(randomized_y_unif, to = c(0 + .Machine$double.eps, 1 - .Machine$double.eps))
    # Convert the y-values back to their original distribution
    # Will convert from uniform distirbution back to original distribution in next step
    randomized_y[,y_index] = randomized_y_unif
  }
  
  # Convert the y-values to uniform distribution to convert back to 
  randomized_y = undo_transform_CE_data(uniform_dist_data_info, randomized_y)
  
  
  # put the names for the response back
  colnames(randomized_y) = y_names
  
  randomized_data = cbind(original_x, randomized_y)
  
  return(randomized_data)
}


