#' Thomas Johnson III
#' 12-22-2024
#' Job Submission Script
#' Runs Model 1 Simulations for GADP for various simulation model Parameters

#' Encountering issues because they do not have the R packages I have starting with DoSNOW.

library(tidyverse)
library(doFuture)
library(future.callr)

# Test increasing number of Model predictors.


# set cores for Longleaf
# One core to handle retrieving and organizing the data
core_alloc = min(parallelly::availableCores(omit = 1), 255)


produce_csv = function(created_object, base_str, file_name_str)
{
  write_csv(created_object, paste0(base_str, file_name_str))
  return(file_name_str)
}

produce_csv_ver_2 = 
  function(created_object, base_str, file_name_str)
  {
    tabular_results = created_object$output
    write_csv(tabular_results, paste0(base_str, file_name_str))
    return(file_name_str)
  }

# Perfrom GADP ---------------------------

gadp_func_ver_2 = function(unprotected_data, target_indices, theta = 0.4, dat.seed = 12)
{
  set.seed(dat.seed)
  
  unprotected_data = as.matrix(unprotected_data)
  dat_means = colMeans(unprotected_data)
  # means of the target variable
  target_means = dat_means[target_indices]
  # means of the auxiliary variable
  auxiliary_means = dat_means[-target_indices]
  # Covariance of the target variable
  target_covariance = cov(unprotected_data)[target_indices,target_indices]
  print("target covariance")
  print(target_covariance)
  
  # Auxiliary variables covariance
  aux_covariance = cov(unprotected_data[,-target_indices])
  
  target_aux_covariance = as.matrix(Matrix::nearPD(cov(unprotected_data))$mat)
  target_aux_covariance = 
    target_aux_covariance[-c(target_indices),
                          c(target_indices)]
  
  # print("auxiliary covariance matrix")
  # print(dim(target_aux_covariance))
  # 
  # print("target covariance matrix")
  # print(dim(target_covariance))
  
  unprotected_cov = cov(unprotected_data)
  
  unprotected_means = colMeans(unprotected_data)
  
  
  
  # random_variates = MASS::mvrnorm(mu = c(target_means, auxiliary_means),
  #                                 Sigma = )
  
  new_y_variates_covariance = unprotected_cov[target_indices,target_indices] * theta^2
  if(length(new_y_variates_covariance) == 1)
  {
    new_y_variates_covariance = cbind(t(unprotected_cov[target_indices, -target_indices]), 
                                      c(new_y_variates_covariance))
  }
  else{
    new_y_variates_covariance = cbind(unprotected_cov[target_indices, -target_indices], 
                                      new_y_variates_covariance)
  }
  
  
  # print("Randomized y covariance dimensions")
  # print(dim(new_y_variates_covariance))
  # 
  # print("target covariance")
  # print(target_covariance)
  # print(dim(target_covariance))
  # print("covariance for new variables")
  # print(new_y_variates_covariance)
  # print("covarinace for unprotected variables")
  # print(unprotected_cov)
  
  print(new_y_variates_covariance)
  print(dim(new_y_variates_covariance))
  print(dim(solve(unprotected_cov)))
  
  right_term = new_y_variates_covariance %*%
    solve(unprotected_cov) %*% t(new_y_variates_covariance)
  
  print(right_term)
  
  right_term = as.matrix(Matrix::nearPD(right_term)$mat)
  
  # print(right_term)
  # 
  # print("check positive definite")
  #   print(matrixcalc::is.positive.definite(right_term))
  #   print(matrixcalc::is.positive.definite(target_covariance))
  
  
  new_covariance = target_covariance - right_term
  print(new_covariance)
  new_covariance = as.matrix(Matrix::nearPD(new_covariance)$mat)
  print(new_covariance)
  
  # print("check positive definite, new covaraince")
  # print(matrixcalc::is.positive.definite(new_covariance))
  
  
  randomized_y = matrix(rep(0, nrow(unprotected_data) * length(target_indices)), nrow = nrow(unprotected_data), byrow = TRUE)
  
  for( index in 1:nrow(unprotected_data))
  {
    new_means = target_means + new_y_variates_covariance %*%
      solve(unprotected_cov) %*% (unprotected_data[index,]- unprotected_means)
    
    new_means = t(new_means)
    
    new_rand_vars = MASS::mvrnorm(n = 1, mu = new_means, Sigma =  new_covariance)
    
    
    
    randomized_y[index,] = new_rand_vars
  }
  
  # print(new_covariance)
  
  colnames(randomized_y) = paste0("y_", 1:length(target_indices))
  
  return(randomized_y)
  
}

# Generate Model 3 ----------------
# Model 3 from Scornet Random FOrest SUbsampling Model 3

correlate_unifrom_m.3_p_ver_1 <- function(N = 1000, rho = 0.1, max_pred = 50, dat.seed = 90,
                                          noise_1 = 0.2, noise_2 = noise_1 * 2)
{
  set.seed(dat.seed)
  
  
  # Generating the data using mu vector and variance-covariance matrix
  sigma.mat = diag(max_pred) + ifelse(diag(max_pred) == 0, rho , 0)
  x_mat_model = apply(
    MASS::mvrnorm(N, mu = rep(0,max_pred), Sigma = sigma.mat, tol = 1),
    2, pnorm)
  
  
  colnames(x_mat_model) <- paste0("X", 1:max_pred)
  # Account for extra predictors, for generating model 1's y
  
  # error for the model
  error = rnorm(N, 0, noise_1)
  error_2 = rnorm(N, 0, noise_2)
  # sine functions, and exponential function
  # Add one to index as the intercept is included
  model_3.rho = -sin(2 * x_mat_model[,1]) + x_mat_model[,2]^2 + x_mat_model[,3] -
    exp( - x_mat_model[,4])
  
  # Generate y with smaller noise
  y_1 <- model_3.rho + error
  
  # Model 1 with more noise
  y_2 <- model_3.rho + error_2
  
  m.3.2_model_rho = data.frame(x_mat_model, y_1, y_2)
  
  return(m.3.2_model_rho)
}


# SUpport Functions for locating Variables -------------

get_variables = function(target_variable_prefix = "y", variable_names)
{
  target_variables = sapply(variable_names, str_detect, 
                            pattern = target_variable_prefix)
  target_variable_indices = which(target_variables == TRUE)
  predictor_variables = variable_names[-target_variable_indices]
  target_and_predictor_vars = list("responses" = variable_names[target_variable_indices],
                                   "predictors" = predictor_variables)
  return(target_and_predictor_vars)
}



sdc_formula_generation = function(target_variables, predictor_variables)
{
  # get the response part of the formula
  response_part = paste0(target_variables,collapse = " + ")
  
  # Get the predictor part of the formula
  predictors_part = paste0(predictor_variables,collapse = " + ")
  
  # Get the complete formula
  complete_formula = formula(paste0(response_part, " ~ ", predictors_part, collapse= ""))
  return(complete_formula)
}

# Evaluation Function -----------------------
# Code for Performing Machine Learning Evaluations With GADP

estimator_eval_stacked_ensemb_iterative_cv = function( # the dataset
  m_start = 1, # Primarily used for debugging determines which iteration to start from
  m = 1000, # Number of Monte Carlo simulations
  samp_n = c(500, 1000, 2000), # sample sizes to be used for training set
  cores = 2, # Number of CPU threads/ logical processors to use
  response_prefix = "y", # prefix for response variables
  acquire_variable_names = get_variables, #Function for getting the variables
  perturbation_func = gadp_func, # CGADP or GADP function
  theta_value = 0.4, # privacy parameter value
  model_generating_function, # function to generate the model
  testing_data_seed = 30000, # Seed for the testing set
  model_predictor_quant = 10, # Quantity of the predictors for the model
  model_rho = 0.2, # correlation of predictors in the model
  model_noise = 0.2, # Noise to be used in the model
  testing_n = 500) # Size of the testing set
{
  # Cross-Validation Funcitons -----------
  # XGBoosted Trees
  xgbt_cv = function(x_pred, y_resp, 
                     parameter_grid = list("nrounds" = c(10,20, 40),
                                           "max_depth" = c(2,4),
                                           "eta" = c(0.2, 0.4, 0.6),
                                           "gamma" = c(0, 0.5, 1),
                                           "colsample_bytree" = c(0.2, 0.5, 1),
                                           "min_child_weight" = c(0.5, 1),
                                           "subsample" = c(0.4, 0.5, 0.6)),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    xgbt_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    xgbt_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "xgbTree", tuneGrid = param_grid,
                            trControl = xgbt_control,
                            nthread = 1,
                            verbose = 0)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_xgbt_obj = xgbt_obj$finalModel
    return(final_xgbt_obj)
  }
  
  # Random forest
  ranger_cv = function(x_pred, y_resp, num_trees = 500,
                       frac_pred = c(1/3, 1/4), 
                       min_node_size = c(5,10),
                       cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = as.matrix(expand_grid(num_trees, frac_pred, min_node_size))
    # Set up folds to be used in cross-validation
    fold_ids = split(1:nrow(x_pred), 1:cv_folds)
    # Vector of average mse for each combination of hyperparameters
    avg_mse_vals = numeric(nrow(param_grid))
    # Loop through each set of hyperparameters
    for(param_index in 1:nrow(param_grid))
    {
      # Vector to store mses per each test fold
      cv_mse_vec = numeric(cv_folds)
      # Cycle through each potential test fold
      for(fold_index in 1:cv_folds)
      {
        # Folds that will be used for training
        cv_sample = unsplit(fold_ids[-fold_index], rep(1:(cv_folds), cv_folds-1))
        # Fold that will be used for test
        cv_test = unsplit(fold_ids[-fold_index], 1:(cv_folds))
        # COnstruct the random forest
        ranger_rf_obj = ranger(x = x_pred[cv_sample,],
                               y = y_resp[cv_sample],
                               mtry = ceiling(ncol(x_pred) * param_grid[param_index, "frac_pred"]),
                               min.node.size = param_grid[param_index, "min_node_size"],
                               num.trees = param_grid[param_index, "num_trees"],
                               num.threads = 1)
        # Get the predictions
        ranger_predictions = predict(ranger_rf_obj, x_pred[cv_sample,])$predictions
        # Calculate the MSE
        mse_inst = 1 / length(cv_sample) * sum((y_resp[cv_sample] - ranger_predictions)^2)
        # Store the MSEs across each fold
        cv_mse_vec[fold_index] = mse_inst
      }
      # Store the average of the MSEs across eeach fold
      avg_mse_vals[param_index] = mean(cv_mse_vec) 
    }
    # Taking the minimum MSE or the first occurrence of the minimum MSE
    min_mse_position = which.min(avg_mse_vals)
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_ranger_obj = ranger(x = x_pred, y = y_resp,
                              mtry = ceiling(ncol(x_pred) * param_grid[min_mse_position, "frac_pred"]),
                              min.node.size = param_grid[min_mse_position, "min_node_size"],
                              num.trees = param_grid[min_mse_position, "num_trees"],
                              num.threads = 1)
    return(final_ranger_obj)
  }
  
  # Neural Network
  nnet_cv = function(x_pred, y_resp, 
                     parameter_grid = list("size" = c(3, 5, 10, 20),
                                           "decay" = c(0.001, 0.01, 0.1) ),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    nnet_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    nnet_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "nnet", tuneGrid = param_grid,
                            trControl = nnet_control,
                            maxit = 500,
                            linout=FALSE,
                            trace=FALSE)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_nnet_obj = nnet_obj$finalModel
    return(final_nnet_obj)
  }
  
  # generate the testing data
  testing_model_data = model_generating_function(N = testing_n, rho = model_rho, 
                                                 max_pred = model_predictor_quant, 
                                                 dat.seed = testing_data_seed,
                                                 noise_1 = model_noise)
  
  # Getting variable names:
  all_variable_names = colnames(testing_model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  # Variables serving as the predictors
  predictor_vars = organized_vars$predictors
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  # Create the cluster (doParallel, doSNOW)
  # sim_cluster = makeCluster(cores)
  # Register the cluster
  # registerDoFuture()
  
  # Use futures with callr to bypass R connections limits
  # plan(multisession, workers = core_alloc)
  plan(callr, workers = core_alloc)
  
  # don't use %<-% as it will throw an error.
  # Store the results in a matrix
  results <- foreach(samp_size_index = 1:length(samp_n), .combine = "rbind", 
                    .options.future=list(packages = c("caret","glmnet", "ranger", 
                                  "nnet", "e1071", "xgboost", "dplyr",
                                  "tidyr", "caret")),
                    .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dofuture% {
              
              # setting the seed
              dat.seed = iteration + samp_n[samp_size_index]
              set.seed(dat.seed)
              
              # Training dataset
              # Generated with the same parameters sxcluding the seed
              training_model_data = 
                model_generating_function(N = samp_n[samp_size_index], 
                                          rho = model_rho, 
                                          max_pred = model_predictor_quant, 
                                          dat.seed = dat.seed,
                                          noise_1 = model_noise)
              
              
              # Performing SDC GADP method
              # Testing data
              privatized_responses_testing = 
                perturbation_func(unprotected_data = testing_model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              # Training data
              privatized_responses_training = 
                perturbation_func(unprotected_data = training_model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              # Combining the training and the test datasets
              privatized_responses = 
                rbind(privatized_responses_training,
                      privatized_responses_testing)
              
              model_data = rbind(training_model_data, 
                                 testing_model_data)
              
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,predictor_vars]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the training set form the df, first 1:n is training, n+1 
              # onwards is the testing data set. 
              drawing = 1:samp_n[samp_size_index]
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              x_dat_holdout = cbind(rep(1, nrow(model_data[-drawing,])),
                                    as.matrix(model_data)[-drawing,predictor_vars])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after gapd
              x_dat = as.matrix(cbind(rep(1, nrow(sample_private_data[,predictor_vars])),
                                      sample_private_data[,predictor_vars]))
              
              ### GAPD Used----
              
              # Linear Regression on data after gapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              
              # LASSO on data after gapd
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,predictor_vars]),
                y = current_response[drawing])
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,predictor_vars]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,predictor_vars]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                   lasso_pred)
              
              ### Neural Net Regression on data after gapd -------------
              
              
              gapd_nnet = nnet_cv(x_pred = x_dat[,predictor_vars], y_resp = current_response[drawing])
              
              
              gapd_nnet_reg_pred = predict(gapd_nnet,x_dat_holdout[,predictor_vars])
              
              # training data predictions
              training_gapd_nnet_reg_pred = predict(gapd_nnet,x_dat[,predictor_vars])
              
              # training data MSE
              training_gapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_nnet_reg_pred)
              
              
              # testing data MSE
              gapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                           gapd_nnet_reg_pred)
              
              # SVM Radial Regression on data after gapd
              gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                              gamma = seq(0.001,0.1,length.out = 2), 
                                              cost = seq(0.1,1,length.out = 2),
                                              kernel = c("radial"))
              
              gapd_radial_svm = gapd_radial_svm_tune$best.model
              
              gapd_radial_svm_reg_pred = predict(gapd_radial_svm,x_dat_holdout[,predictor_vars])
              
              # Get the training predictions of the radial svm
              training_gapd_radial_svm_reg_pred =
                predict(gapd_radial_svm,x_dat[,predictor_vars])
              
              # Get the training mse for radial svm
              training_gapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_radial_svm_reg_pred)
              
              # Testing data MSE
              gapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after gapd
              gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                              
                                              cost = seq(0.1,1,length.out = 2),
                                              kernel = c("linear"))
              
              gapd_linear_svm = gapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_gapd_linear_svm_reg_pred =
                predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
              
              # Training MSE for Linear SVM
              training_gapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              gapd_linear_svm_reg_pred = predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
              
              # Testing data MSE
              gapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                            gamma = seq(0.001,0.1,length.out = 2), 
                                            cost = seq(0.1,1,length.out = 2),
                                            degree = seq(1,2,length.out=2),
                                            kernel = c("polynomial"),
                                            cross = 10)
              
              gapd_poly_svm = gapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat[,predictor_vars])
              
              # Training data MSE polynomial SVM
              training_gapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat_holdout[,predictor_vars])
              
              # MSE for testing dataset for polynomial kernel SVM
              gapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                               gapd_poly_svm_reg_pred)
              
              # Random forest on data after gapd
              num_x_vars = length(predictor_vars)
              
              ranger_rf_regress_model = ranger_cv( 
                x_pred = as.matrix(sample_private_data[,predictor_vars]),
                y_resp = current_response[drawing])
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,predictor_vars])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,predictor_vars])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after gapd
              xgbt_regress_model = xgbt_cv( 
                x_pred = as.matrix(sample_private_data[,predictor_vars]),
                y_resp = current_response[drawing])
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,predictor_vars]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,predictor_vars]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                  xgbt_pred)
              
              ##
              ## Before GAPD (On original data) ------
              print(" running checkpoint 1")
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = as.vector(x_dat %*%orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout %*% orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(sample_unprotected_data[,predictor_vars]),
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,predictor_vars]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,predictor_vars]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_gapd_nnet = nnet_cv(x_pred = x_dat[,predictor_vars], 
                                     y_resp = sample_orig_data_response)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat[,predictor_vars])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat_holdout[,predictor_vars])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              
              
              no_gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                                 y = sample_orig_data_response,
                                                 gamma = seq(0.001,0.1,length.out = 2), 
                                                 cost = seq(0.1,1,length.out = 2),
                                                 
                                                 kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_gapd_radial_svm = no_gapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat[,predictor_vars])
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat_holdout[,predictor_vars])
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                                 y = sample_orig_data_response, 
                                                 cost = seq(0.1,1,length.out = 2),
                                                 kernel = c("linear"))
              # The Linear SVM best tuned model
              no_gapd_linear_svm = no_gapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat[,predictor_vars])
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat_holdout[,predictor_vars])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              
              no_gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                               y = sample_orig_data_response,
                                               gamma = seq(0.001,0.1,length.out = 2), 
                                               cost = seq(0.1,1,length.out = 2),
                                               degree = seq(1,2,length.out=2),
                                               kernel = c("polynomial"),
                                               cross = 10)
              # Get the best tuned polynomial SVM
              no_gapd_poly_svm = no_gapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat[,predictor_vars])
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat_holdout[,predictor_vars])
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before gapd
              
              ranger_rf_model_orig_data = ranger_cv( 
                x_pred = as.matrix(sample_unprotected_data[,predictor_vars]),
                y_resp = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,predictor_vars])$predictions
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,predictor_vars])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before gapd
              xgbt_regress_model_orig_data =
                xgbt_cv( x_pred = as.matrix(sample_unprotected_data[,predictor_vars]),
                         y_resp = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[drawing,predictor_vars]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,predictor_vars]))
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         xgbt_pred_orig_data)
              
              
              ###
              ### Stacked Ensemble Equal Weighting ----
              # GAPD
              avg_stacked_ensemble_predictions_gapd = (linear_reg_pred  + 
                                                         lasso_pred  +  gapd_nnet_reg_pred  + 
                                                         gapd_radial_svm_reg_pred  + 
                                                         gapd_linear_svm_reg_pred  + 
                                                         gapd_poly_svm_reg_pred  +
                                                         ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_predictions_gapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_gapd = 
                (linear_reg_unprivatized_pred  + 
                   lasso_pred_orig_data  + 
                   nnet_reg_unprivatized_pred + 
                   radial_svm_reg_unprivatized_pred  + 
                   linear_svm_reg_unprivatized_pred  + 
                   poly_svm_reg_unprivatized_pred  +
                   ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_pred_no_gapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_gapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_gapd_nnet_reg_mse, training_gapd_radial_svm_reg_mse,
                  training_gapd_linear_svm_reg_mse, 
                  training_gapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_gapd_weighting = 
                (training_set_mses_gapd / sum(training_set_mses_gapd))^(-1)
              prop_var_exp_gapd_weighting = prop_var_exp_gapd_weighting / 
                sum(prop_var_exp_gapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_gapd_pred = 
                linear_reg_pred * prop_var_exp_gapd_weighting[1] +
                lasso_pred * prop_var_exp_gapd_weighting[2] + 
                gapd_nnet_reg_pred * prop_var_exp_gapd_weighting[3] + 
                gapd_radial_svm_reg_pred * prop_var_exp_gapd_weighting[4] + 
                gapd_linear_svm_reg_pred * prop_var_exp_gapd_weighting[5] + 
                gapd_poly_svm_reg_pred * prop_var_exp_gapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_gapd_weighting[7] + 
                xgbt_pred * prop_var_exp_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_gapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_gapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_gapd_weighting = 
                (training_set_mses_no_gapd / sum(training_set_mses_no_gapd))^(-1)
              prop_var_exp_no_gapd_weighting = prop_var_exp_no_gapd_weighting /
                sum(prop_var_exp_no_gapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_gapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_gapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_gapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_no_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_no_gapd_pred)
              
              
              # Compile MSEs into a vector
              gapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, gapd_nnet_reg_mse,
                  gapd_radial_svm_reg_mse, gapd_linear_svm_reg_mse,
                  gapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_gapd_mse, 
                  prop_var_stacked_ens_gapd_mse)
              
              no_gapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_gapd_mse, 
                  prop_var_stacked_ens_no_gapd_mse)
              
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "training_size", 
                                   "testing_size",
                                   "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","gadp", "theta", "rho")
              
              
              # construct the dataframe of results
              estimator_reg_df = data.frame(rbind(c(dat.seed,iteration, samp_n[samp_size_index], testing_n,
                                                    response_index, gapd_mse_vector, 1, theta_value, model_rho)))
              
              # Adding results for predictive models without GAPD
              estimator_reg_df[2,] = rbind(c(dat.seed, iteration, 
                                             samp_n[samp_size_index], testing_n,
                                             response_index, no_gapd_mse_vector, 0, theta_value, model_rho))
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              
              return(estimator_reg_df)
            }
  
  # stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  return(results)
}


# Parameters for the Simualtion Models ------
#' Initializing the parameters to be able to construct the simulation models.

# Sample sizes
n_sizes = c(500, 1000, 2000)

# Correlation amongst the predictors
rho_params = seq(0.2,0.6, by=0.2)

# Privacy Parameter, lower is more privacy and less data utility
var_theta = seq(0.2,0.8, by=0.2)

# Quantity of Predictors
# Will start to increase compute time for the SVMs more and more.
# particularly the Polynomial SVM
# Getting unsaved results for predictors = 10
predictor_quant = c(10, 50, 100)

# Number of simulations to run
m_sims = 1000

# Get the start and ends of the simulations
# Use in case there are memory issues
# sim_start = c(1, 201, 401, 601, 801)
# sim_end = c(200, 400, 600, 1000)

# Get all combinations of simulation parameters
# Edit as necessary as the simulations likely will not finish in one run
param_df = expand_grid(rho_params, var_theta, predictor_quant)
param_df = param_df %>% arrange(predictor_quant, var_theta, rho_params)


# Set the working Directory
#setwd(here::here())



base_dir = "~/" # Your base directory goes here (where it will reference to store results)

setwd(base_dir)

# Setup broader data directory 
initial_dat_dir = paste0(base_dir,"/Data/")

if(dir.exists(initial_dat_dir) == FALSE)
{
  dir.create(initial_dat_dir)
}

# Setup data directory 
data_dir = paste0(initial_dat_dir,"tuned-sims/")

if(dir.exists(data_dir) == FALSE)
{
  dir.create(data_dir)
}

# Data Dirctory for Model 3
model_3_dir = paste0(data_dir, "Model-3/")

if(dir.exists(model_3_dir) == FALSE)
{
  dir.create(model_3_dir)
}

for(row_index in 1:nrow(param_df))
{
  results_df = estimator_eval_stacked_ensemb_iterative_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
                                                            m = m_sims, # end of portion of simulations
                                          samp_n = n_sizes, # sample sizes to be used for training set
                                          cores = core_alloc, # Number of CPU threads/ logical processors to use
                                          response_prefix = "y", # prefix for response variables
                                          acquire_variable_names = get_variables, #Function for getting the variables
                                          perturbation_func = gadp_func_ver_2, # CGADP or GADP function
                                          theta_value = param_df$var_theta[row_index], # privacy parameter value
                                          model_generating_function = correlate_unifrom_m.3_p_ver_1, # function to generate the model
                                          testing_data_seed = 30000, # Seed for the testing set
                                          model_predictor_quant = param_df$predictor_quant[row_index], # Quantity of the predictors for the model
                                          model_rho = param_df$rho_params[row_index], # correlation of predictors in the model
                                          model_noise = 0.2, # Noise to be used in the model
                                          testing_n = 500 # size of testing set
                                          )
  # string for the rho value
  rho_string = str_replace(paste0(param_df$rho_params[row_index]), "0.", "0_")
  # String for the privacy parameter value
  var_theta_string = str_replace(paste0(param_df$var_theta[row_index]), "0.", "0_")
  # String for the qunatity of predictors
  predictor_string = paste0(param_df$predictor_quant[row_index])
  # Name of the results
  results_file_name = paste0("model-3-gadp-rho-",
                        rho_string, "-theta-",
                        var_theta_string, 
                        "-pred-", predictor_string,".csv")
  
  produce_csv(created_object= results_df, base_str = model_3_dir, file_name_str = results_file_name)
  
  remove(results_df)
  remove(results_file_name)
}
# CLose the R session when COde is done
quit(save = "no")


