#' Thomas Johnson III
#' 11-25-2023
#' Iteratively generated model construction code
#' 




correlate_uniform_m.1_p_ver_2 <- function(N = 1000, rho = 0.1, max_pred = 50, dat.seed = 90,
                                          noise_1 = 0.2, noise_2 = noise_1 * 2)
{
  # set the seed
  set.seed(dat.seed)
  base_pred = 8
  num_pred = max_pred - base_pred # extra predictors that are noise
  
  
  # Generating the data using mu vector and variance-covariance matrix
  sigma.mat = diag(max_pred) + ifelse(diag(max_pred) == 0, rho , 0)
  x_mat_model = apply(cbind(rep(1,N), 
                            MASS::mvrnorm(N, mu = rep(0,max_pred), Sigma = sigma.mat, tol = 1)), 
                      2, pnorm)
  colnames(x_mat_model) <- paste0("X", 0:max_pred)
  # Account for extra predictors, for generating model 1's y
  coeff.mat_m.1 = matrix(c(1, 0, 1, 0, 1.5, 0, 0, 0, 1, rep(0,num_pred)), 
                         nrow =max_pred+1, ncol = 1, byrow = TRUE)
  
  # error for the model
  error = rnorm(N, 0, noise_1)
  error_2 = rnorm(N, 0, noise_2)
  
  model_1.rho =x_mat_model %*% coeff.mat_m.1
  
  # Model 1 from McConvill 2017 Dissertation
  y_1 <- model_1.rho + error
  # 
  # # Account for extra predictors generating model 2's y
  # # Have account for the lack of one predictor
  # coeff.mat_m.2 = matrix(c(1.5, 0, 4, 0, 0, 8, 0, 0, rep(0, num_pred + 1)), 
  #                    nrow =max_pred+1, ncol = 1, byrow = TRUE)
  # model_2.rho = x_mat_model %*% coeff.mat_m.2
  # # Model 2 from McConville 2017 Dissertation, modified form to get positive definite matrix
  
  # Model 1 with more noise
  y_2 <- model_1.rho + error_2
  
  m.1.2_model_rho = data.frame(x_mat_model, y_1, y_2)
  # Remove the intercept term
  m.1.2_model_rho = m.1.2_model_rho[,-1]
  return(m.1.2_model_rho)
}

# Model 2 Simulation Code
# Used to iteratively generate Model 2
correlate_uniform_m.2_p_ver_1 <- function(N = 1000, rho = 0.1, max_pred = 50, dat.seed = 90,
                                          noise_1 = 0.2)
{
  set.seed(dat.seed)
  base_pred = 8
  num_pred = max_pred - base_pred # extra predictors that are noise
  
  noise_2 = 2 * noise_1
  
  
  # Generating the data using mu vector and variance-covariance matrix
  sigma.mat = diag(max_pred) + ifelse(diag(max_pred) == 0, rho , 0)
  x_mat_model = cbind(
    MASS::mvrnorm(N, mu = rep(0,max_pred), Sigma = sigma.mat, tol = 1))
  
  
  colnames(x_mat_model) <- paste0("X", 1:max_pred)
  # Account for extra predictors, for generating model 1's y
  
  
  # error for the model
  error = rnorm(N, 0, noise_1)
  error_2 = rnorm(N, 0, noise_2)
  
  model_2.rho = 1+ 1.5 * x_mat_model[,"X1"]- 0.5 * x_mat_model[,"X2"] + 4 * (x_mat_model[,"X4"] < 0.5) -
    2 * (x_mat_model[,"X4"] >= 0.5) + 1.5 * (x_mat_model[,"X5"] < 0.25) + x_mat_model[,"X6"]
  
  y_1 <- model_2.rho + error
  # 
  
  # Model with more noise
  y_2 <- model_2.rho + error_2
  
  m.2.2_model_rho = data.frame(x_mat_model, y_1, y_2)
  # Remove the intercept term
  m.2.2_model_rho = m.2.2_model_rho
  return(m.2.2_model_rho)
}

# Used to explore the separate parts of the mixture
# correlate_unifrom_m.2_p_break_part_1 <- function(N = 1000, rho = 0.1, max_pred = 50, dat.seed = 90,
#                                           noise_1 = 0.2)
# {
#   set.seed(dat.seed)
#   base_pred = 8
#   num_pred = max_pred - base_pred # extra predictors that are noise
#   
#   noise_2 = 2 * noise_1
#   
#   
#   # Generating the data using mu vector and variance-covariance matrix
#   sigma.mat = diag(max_pred) + ifelse(diag(max_pred) == 0, rho , 0)
#   x_mat_model = cbind(
#     MASS::mvrnorm(N, mu = rep(0,max_pred), Sigma = sigma.mat, tol = 1))
#   
#   
#   colnames(x_mat_model) <- paste0("X", 1:max_pred)
#   # Account for extra predictors, for generating model 1's y
#   
#   
#   # error for the model
#   error = rnorm(N, 0, noise_1)
#   error_2 = rnorm(N, 0, noise_2)
#   
#   model_2.rho = 1+ 1.5 * x_mat_model[,"X1"]- 0.5 * x_mat_model[,"X2"] + 4 * 
#     (x_mat_model[,"X4"] < 0.5) + 1.5 * (x_mat_model[,"X5"] < 0.25) + x_mat_model[,"X6"]
#   
#   y_1 <- model_2.rho + error
#   # 
#   
#   # Model with more noise
#   y_2 <- model_2.rho + error_2
#   
#   m.2.2_model_rho = data.frame(x_mat_model, y_1, y_2)
#   # Remove the intercept term
#   m.2.2_model_rho = m.2.2_model_rho
#   return(m.2.2_model_rho)
# }
# 
# # Used to explore the mixture
# correlate_unifrom_m.2_p_break_part_2 <- function(N = 1000, rho = 0.1, max_pred = 50, dat.seed = 90,
#                                                  noise_1 = 0.2)
# {
#   set.seed(dat.seed)
#   base_pred = 8
#   num_pred = max_pred - base_pred # extra predictors that are noise
#   
#   noise_2 = 2 * noise_1
#   
#   
#   # Generating the data using mu vector and variance-covariance matrix
#   sigma.mat = diag(max_pred) + ifelse(diag(max_pred) == 0, rho , 0)
#   x_mat_model = cbind(
#     MASS::mvrnorm(N, mu = rep(0,max_pred), Sigma = sigma.mat, tol = 1))
#   
#   
#   colnames(x_mat_model) <- paste0("X", 1:max_pred)
#   # Account for extra predictors, for generating model 1's y
#   
#   
#   # error for the model
#   error = rnorm(N, 0, noise_1)
#   error_2 = rnorm(N, 0, noise_2)
#   
#   model_2.rho = 1+ 1.5 * x_mat_model[,"X1"]- 0.5 * x_mat_model[,"X2"]  -
#     2 * (x_mat_model[,"X4"] >= 0.5) + 1.5 * (x_mat_model[,"X5"] < 0.25) + 
#     x_mat_model[,"X6"]
#   
#   y_1 <- model_2.rho + error
#   # 
#   
#   # Model with more noise
#   y_2 <- model_2.rho + error_2
#   
#   m.2.2_model_rho = data.frame(x_mat_model, y_1, y_2)
#   # Remove the intercept term
#   m.2.2_model_rho = m.2.2_model_rho
#   return(m.2.2_model_rho)
# }

# Model 3 from Scornet Model 3
# Used to iteratively generate Model 3
correlate_uniform_m.3_p_ver_2 <- function(N = 1000, rho = 0.1, max_pred = 50, dat.seed = 90,
                                          noise_1 = 0.2, noise_2 = noise_1 * 2)
{
  set.seed(dat.seed)
  
  
  # Generating the data using mu vector and variance-covariance matrix
  sigma.mat = diag(max_pred) + ifelse(diag(max_pred) == 0, rho , 0)
  x_mat_model = apply( 
                            MASS::mvrnorm(N, mu = rep(0,max_pred), Sigma = sigma.mat, tol = 1), 
                      2, pnorm)
  
  
  colnames(x_mat_model) <- paste0("X", 1:max_pred)
  # Account for extra predictors, for generating model 1's y
  
  # error for the model
  error = rnorm(N, 0, noise_1)
  error_2 = rnorm(N, 0, noise_2)
  # sine functions, and exponential function
  # Add one to index as the intercept is included
  model_3.rho = -sin(2 * x_mat_model[,1]) + x_mat_model[,2]^2 + x_mat_model[,3] -
    exp( - x_mat_model[,4])
  
  # Model 1 from McConvill 2017 Dissertation
  y_1 <- model_3.rho + error
  # 
  # # Account for extra predictors generating model 2's y
  # # Have account for the lack of one predictor
  # coeff.mat_m.2 = matrix(c(1.5, 0, 4, 0, 0, 8, 0, 0, rep(0, num_pred + 1)), 
  #                    nrow =max_pred+1, ncol = 1, byrow = TRUE)
  # model_2.rho = x_mat_model %*% coeff.mat_m.2
  # # Model 2 from McConville 2017 Dissertation, modified form to get positive definite matrix
  
  # Model 1 with more noise
  y_2 <- model_3.rho + error_2
  
  m.3.2_model_rho = data.frame(x_mat_model, y_1, y_2)
  
  return(m.3.2_model_rho)
}
