# Impact of Data Perturbation on ML Techniques]{Impact of Data Perturbation for Statistical Disclosure Control on the Predictive Performance of Machine Learning Techniques README.md

By Thomas Johnson III & Sayed Mostafa

## Dependencies

Packages needed: ("MASS", "ranger", "mixtools", "EnvStats", "xgboost", "e1071", "tidyverse",
"doFuture", "doSNOW", "future.callr", "glmnet", "Matrix", "tidyverse", "knit", "caret",
 "nnet", "foreach", "future", "tables", "here", "knitr", "rpms", "tidyr", "targets", "stringr")
 
 You can install with `install.packages()` or using the "renv" package and the `install()` function.

## Important Assumption to Check

The code using targets assumes that you are in the project directory "SDC-Comparisons-Submission". 

## Important Notes

- Note: When "." is used in this README.md it is referring to the project directory. The directory
where "SDC-Comparisons.Rproj" is located.

- Note: Have only one copy of the project uncompressed in your file system. Previous versions in the same filesystem can cause issues.

### Running Through Targets

If running the simulations via targets, please check the output to make sure the correct directory is being located by the `here()` function. If the wrong path is being used, please change `base_dir` variable to be the path to the project like "path/to/project/SDC-Comparisons-Submission/".

### Running Through Scripts

If running the simulation code through the scripts found in the "SDC-Comparisons-Submission/Code/General-Scripts/" subdirectory, please specify the directory path ("path/to/project/SDC-Comparisons-Submission/") manually for the variable `base_dir`. These scripts assume you are running the code in a remote environment where you are unable to open the R project like you would in RStudio (automatically obtained working directory, etc.). These scripts were used to run the code on an HPC system using SLURM. Similarly, the CE data can be loaded, cleaned, and loaded into the real data example simulations using the "SDC-Comparisons-Submission/Code/CE-data-outlier-removal-cv.R" script. For the "CE-data-outlier-removal-cv.R", change `base_dir` to the directory path for the project ("path/to/project/SDC-Comparisons-Submission/") if the here() function cannot find the path. The here() function can typically locate the project when opened in RStudio.

## Instructions to Run Code

Run the targets file first with `targets::tar_make()`. It should run all simulation results. Please note, this
assumes that you are running the code from a setup where the `here()` can correctly locate the directory of this project. This would assume you can open the R project. Please note that the data for the simulated models is generated via the functions in "./Code/Iterative-Model-Gen-Code.R" and the CE data for the real dataset is obtainable via the R package
'rpms'. The "get_CE_data()" function will perform the necessary data retrieval and cleaning for the CE data.

If there are issues running the real data application on the CE data part of the targets file, you can run 
"CE-data-outlier-removal-cv.R" to get the real data simulations. 

If there is an issue running the simulations Models sections of the targets file, you
can go to the General-Scripts sub directory inside the Code sub directory ("./Code/General-Scripts/") and run the individual scripts to get the results. Please specify `base_dir` 
with a path that the code can store the results in. The scripts in the General-Scripts folder should be self-contained, with necessary functions included. The general scripts will not install the packages that are necessary on their own own. The general scripts used future and doFuture for parallelization.

## Getting Results From CSVs

The results can be loaded within two Quarto documents. These Quarto documents were used to produce the results seen in the manuscript.

To get results for the CE data, please run the "CE-Real-Data-Tables-fincbtax-salaryx-tuned-sims.qmd" document. 
It is in the "Reports" folder "./Code/Reports/CE-Real-Data-Tables-fincbtax-salaryx-tuned-sims.qmd".

To get results for the simulation models, please run the 
"More-tuned-SDC-GAPD-CGAPD-Comparisons-01-29-2025.qmd" document. It is in the "Reports"" folder: "./Code/Reports/More-tuned-SDC-GAPD-CGAPD-Comparisons-01-29-2025.qmd".

## Files

- "./targets.R": Runs the simulations in a targets pipeline to further reinforce reproducibility and provide safeguards in case the code is interrupted midrun. 
The targets pipeline should be able restart from the last computed object in the pipeline if interrupted. Has multiple dependencies for modularity (see the `source()` function usage in 
"targets.R" file). It has not been tested on an HPC system. Only on a Windows and Linux desktops/laptops.

- "./SDC-Comparisons.Rproj": The R project file that can be opened in RStudio and similar software to allow the project to run with correct directory structure using
the `here()` function.

- "./Code/Basic-Funcs.R": A script containing basic functions that are used across analysis perfromed in the Quarto documents and the targets.R file. It is a dependency
across multiple aspects of the project.

- "./Code/CE-data-cleaning.R": A script containing the functions needed to clean the CE data for the real data application. It is a dependency for targets.R.

- "./Code/CE-data-outlier-removal-cv.R": A script for running the real data example if there is an issue running it in targets.R.
It can also be used to run the code in an HPC system assuming the directories are properly setup. Its dependencies are "./Code/Basic-Funcs.R" and
"./Code/Estimator-Evaluation.R".

- "./Code/Estimator-Evaluation.R": A script containing the functions necessary to perform the simulation runs with the predictive models
for both the simulation data and the real data example. Immediate dependency on "./Code/GADP-from-scratch.R".

- "./Code/GADP-from-scratch.R": A script for the GADP and CGADP functions of the code. Some of the functions can be used generally and some of the functions,
due to the assumptions and requirements of the the method (in particular CGADP), can only be used for a particular set of simulation models or the real data example.

- "./Code/Iterative-Model-Gen-Code.R": Code necessary to generate the simulation data. It is a critical dependency for the simulation data runs. If you want to generate examples of the 
simulated data, these are the functions that would be used.

- "./Reports/CE-Real-Data-Tables-fincbtax-salaryx-tuned-sims.qmd": The Quarto document necessary for reproducing the tables for the real data example. Note that the functions used do not create perfect
relicas of the tables and construct duplicate rows. Certain tables were manually combined in Latex as needed.

- "./Reports/More-tuned-SDC-GAPD-CGAPD-Comparisons-01-29-2025.qmd": The Quarto document necessary for doing analysis on the simulated data, including getting the plots and tables. Note that
the Latex tables may not be perfect replicas of those of in the manuscript and manual edits were necessary.

- "./Reports/refs.bib": Necessary for citations to appear correctly in the Quarto documents.

- "./General-Scripts/Model-1-CGADP-General-Script.R": A self-contained script for running the results for Model 1. This script would be preferred to run the Monte Carlo simulations for Model 1 under 
CGADP as a job in an HPC system. Please properly configure `base_dir` and when finished copy the results to the correct model folder in "./Data/tuned-sims/".

- "./General-Scripts/Model-1-GADP-General-Script.R": A self-contained script for running the results for Model 1. This script would be preferred to run the Monte Carlo simulations for Model 1 under 
GADP as a job in an HPC system. Please properly configure `base_dir` and when finished copy the results to the correct model folder in "./Data/tuned-sims/".

- "./General-Scripts/Model-2-CGADP-General-Script.R": A self-contained script for running the results for Model 2. This script would be preferred to run the Monte Carlo simulations for Model 1 under 
CGADP as a job in an HPC system. Please properly configure `base_dir` and when finished copy the results to the correct model folder in "./Data/tuned-sims/".

- "./General-Scripts/Model-2-GADP-General-Script.R": A self-contained script for running the results for Model 2. This script would be preferred to run the Monte Carlo simulations for Model 1 under 
GADP as a job in an HPC system. Please properly configure `base_dir` and when finished copy the results to the correct model folder in "./Data/tuned-sims/".

- "./General-Scripts/Model-3-CGADP-General-Script.R": A self-contained script for running the results for Model 3. This script would be preferred to run the Monte Carlo simulations for Model 1 under 
CGADP as a job in an HPC system. Please properly configure `base_dir` and when finished copy the results to the correct model folder in "./Data/tuned-sims/".

- "./General-Scripts/Model-3-GADP-General-Script.R": A self-contained script for running the results for Model 3. This script would be preferred to run the Monte Carlo simulations for Model 1 under 
GADP as a job in an HPC system. Please properly configure `base_dir` and when finished copy the results to the correct model folder in "./Data/tuned-sims/".

"./Data/cleaned-CE.csv": A cleaned version of the CE data. It can be obtained via running the functions in "./Code/CE-data-cleaning.R".

## Data

The data for the simulation models is generated by the code as stated in the Instructions to Run Code section. To get an example of the simulation model data, you can generate it via the functions in the "./Code/Iterative-Model-Gen-Code.R" file. Since the data is generated iteratively during the simulations, it would be problematic to include over 10,000+ datasets. It would present a problem in terms of storage. Generating the data via the functions "./Code/Iterative-Model-Gen-Code.R" is cleaner and more manageable. The pertrubed versions of the datasets are obtainable via the use of perturbation functions in "GADP-from-scratch.R". Please note that some of the functions only generate a randomized version of Y, as X is not considered sensitive and remains the same in both the perturbed and non-perturbed datasets in this research endeavor.

The CE data is obtainable from the `rpms` package. The code will clean the data as specified in the Instructions to Run Code section. If you want a copy of the CE data that is used please see "./Data/cleaned-CE.csv". This same data is easily obtainable by running the `get_CE_data()` function (after loading the libraries in the "./Code/CE-data-cleaning.R" file) in "./Code/CE-data-cleaning.R".

## Supplementary Document of Manuscript

Contains supplementary tables and plots that were not included within the main manuscript.

